#ifndef c3d_H
#define c3d_H

#ifdef _WINDOWS
#define WINDOWS
#endif

#include "ac_plugin.h"
#include <math.h>

// Planes
#define C3D_PLANE_X (0)
#define C3D_PLANE_Y (1)
#define C3D_PLANE_Z (2)

// Ray-tri intersects
#define C3D_ISECT_SUCCESS        0
#define C3D_ISECT_DEGENERATE_TRI 1
#define C3D_ISECT_NO_INTERSECT   2
#define C3D_ISECT_OUTSIDE_TRI    3

//---------------------------------------------------------------------------------------
// C3d : General 3d functions
class C3d
{

public:

   class CMatrix
   {
   public:
      float m[4][4];
      // Concatenate
      void Concat( CMatrix &m1, CMatrix &m2 )
      {
         for( int col = 0; col < 4; ++col )
         {
            for( int row = 0; row < 4; ++row )
            {
               this->m[col][row] = m1.m[0][row] * m2.m[col][0] +
                                   m1.m[1][row] * m2.m[col][1] +
                                   m1.m[2][row] * m2.m[col][2] +
                                   m1.m[3][row] * m2.m[col][3];
            }
         }
      }
      // Identity matrix
      void BuildIdentity()
      {
         // Identity matrix
         this->m[0][0] = 1;
         this->m[1][0] = 0;
         this->m[2][0] = 0;
         this->m[3][0] = 0;
         this->m[0][1] = 0;
         this->m[1][1] = 1;
         this->m[2][1] = 0;
         this->m[3][1] = 0;
         this->m[0][2] = 0;
         this->m[1][2] = 0;
         this->m[2][2] = 1;
         this->m[3][2] = 0;
         this->m[0][3] = 0;
         this->m[1][3] = 0;
         this->m[2][3] = 0;
         this->m[3][3] = 1;
      }
      // Build an X rotation matrix
      void BuildRotationX( float angle )
      {
         this->BuildIdentity();
         this->m[1][1] =  cos( angle );
         this->m[1][2] = -sin( angle );
         this->m[2][1] =  sin( angle );
         this->m[2][2] =  cos( angle );
      }
      // Build a Y rotation matrix
      void BuildRotationY( float angle )
      {
         this->BuildIdentity();
         this->m[0][0] =  cos( angle );
         this->m[0][2] =  sin( angle );
         this->m[2][0] = -sin( angle );
         this->m[2][2] =  cos( angle );
      }
      // Build a Z rotation matrix
      void BuildRotationZ( float angle )
      {
         this->BuildIdentity();
         this->m[0][0] =  cos( angle );
         this->m[0][1] = -sin( angle );
         this->m[1][0] =  sin( angle );
         this->m[1][1] =  cos( angle );
      }
      // Build an XYZ rotation matrix (ordered rotation)
      void BuildRotationXYZ( Vertex *angle, Point3 *center )
      {
         CMatrix m1;
         m1.BuildRotationX( angle->x );
         CMatrix m2;
         m2.BuildRotationY( angle->y );
         CMatrix m3;
         m3.Concat( m1, m2 );
         m2.BuildRotationZ( angle->z );
         this->Concat( m3, m2 );
         this->m[3][0] = center->x;
         this->m[3][1] = center->y;
         this->m[3][2] = center->z;
      }
   };

   // Calculate the area of a parallelogram
   static float AreaOfParallelogram( Vertex *v1, Vertex *v2, Vertex *v3 )
   {
      // Calculate vector lengths
      Vertex l1, l2, cross;
      l1.x = v2->x - v1->x;
      l1.y = v2->y - v1->y;
      l1.z = v2->z - v1->z;
      l2.x = v3->x - v1->x;
      l2.y = v3->y - v1->y;
      l2.z = v3->z - v1->z;
      // Get the cross product of the vector lengths
      cross.x = pow(    l1.y * l2.z - l2.y * l1.z  , 2 );
      cross.y = pow( -( l1.x * l2.z - l2.x * l1.z ), 2 );
      cross.z = pow(    l1.x * l2.y - l2.x * l1.y  , 2 );
      // Calculate the magnitude
      cross.x = pow( cross.x, 2 );
      cross.y = pow( cross.y, 2 );
      cross.z = pow( cross.z, 2 );
      // Return the area
      return sqrtf( cross.x + cross.y + cross.z );
   }

   // Calculate the area of a triangle
   static float AreaOfTriangle( Vertex *v1, Vertex *v2, Vertex *v3 )
   {
      return ( AreaOfParallelogram( v1, v2, v3 ) * 0.5f );
   }

   // Return the "dominant" plane along which a normal lies
   static int GetDominantPlaneFromNormal( Point3 &norm )
   {
      if( fabs( norm.y ) > fabs( norm.x ) + 0.00005f )
      {
         if( fabs( norm.z ) > fabs( norm.y ) + 0.00005f )
         {
            return C3D_PLANE_Z;
         }
         else
         {
            return C3D_PLANE_Y;
         }
      }
      else
      {
         if( fabs( norm.z ) > fabs( norm.x ) + 0.00005f )
         {
            return C3D_PLANE_Z;
         }
         else
         {
            return C3D_PLANE_X;
         }
      }
   }

   // Cross product
   static Vertex CrossProduct( Vertex &v1, Vertex &v2 )
   {
      Vertex out;
      out.x = v1.y * v2.z - v1.z * v2.y;
      out.y = ( v1.z * v2.x - v1.x * v2.z );
      out.z = v1.x * v2.y - v1.y * v2.x;
      return out;
   }

   // Dot product
   static float DotProduct( Vertex &v1, Vertex &v2 )
   {
      return v1.x * v2.x + v1.y * v2.y + v1.z * v2.z;
   }

   // Intersect a triange with a ray, return the collision point in outI
   static int RayTriangleIntersect( Vertex *raystart, Vertex *raydir, Vertex triangle[3], Vertex *outI )
   {

      // Get triangle edge vectors and plane normal
      Vertex u;
      u.x = triangle[1].x - triangle[0].x;
      u.y = triangle[1].y - triangle[0].y;
      u.z = triangle[1].z - triangle[0].z;
      Vertex v;
      v.x = triangle[2].x - triangle[0].x;
      v.y = triangle[2].y - triangle[0].y;
      v.z = triangle[2].z - triangle[0].z;
      Vertex n = C3d::CrossProduct( u, v );

      Vertex w0;
      w0.x = raystart->x - triangle[0].x;
      w0.y = raystart->y - triangle[0].y;
      w0.z = raystart->z - triangle[0].z;
      float a = -( C3d::DotProduct( n, w0 ));
      float b = C3d::DotProduct( n, *raydir );
      if( fabs( b ) < 0.0000001f )
      {
         // ray is parallel to triangle plane
         if( a == 0 )
         {
            // ray lies in triangle plane
         }
         else
         {
            // ray disjoint from plane
         }
         return C3D_ISECT_DEGENERATE_TRI;
      }

      // get intersect point of ray with triangle plane
      float r = a / b;
      if( r < 0.0 )
      {
         // ray goes away from triangle
         return C3D_ISECT_NO_INTERSECT;
         // for a segment, also test if (r > 1.0) => no intersect
      }

      Vertex dirmag;
      dirmag.x = raydir->x * r;
      dirmag.y = raydir->y * r;
      dirmag.z = raydir->z * r;
      // intersect point of ray and plane
      outI->x = raystart->x + dirmag.x;
      outI->y = raystart->y + dirmag.y;
      outI->z = raystart->z + dirmag.z;

      // is intersect point in triangle
      float uu, uv, vv, wu, wv, D;
      uu = C3d::DotProduct( u, u );
      uv = C3d::DotProduct( u, v );
      vv = C3d::DotProduct( v, v );
      Vertex w;
      w.x = outI->x - triangle[0].x;
      w.y = outI->y - triangle[0].y;
      w.z = outI->z - triangle[0].z;
      wu = C3d::DotProduct( w, u );
      wv = C3d::DotProduct( w, v );
      D = uv * uv - uu * vv;

      // get and test parametric coords
      float s, t;
      s = ( uv * wv - vv * wu ) / D;
      if( s < 0.0 || s > 1.0 )
      {
         // I is outside T
         return C3D_ISECT_OUTSIDE_TRI;
      }
      t = ( uv * wu - uu * wv ) / D;
      if( t < 0.0 || ( s + t ) > 1.0 )
      {
         // I is outside T
         return C3D_ISECT_OUTSIDE_TRI;
      }

      return C3D_ISECT_SUCCESS;

   }

   // Calculate barycentric coordinates for a point
   static void BarycentricCoords( Vertex &point, Vertex &x1, Vertex &x2, Vertex &x3, Vertex &bcoords, float *baryover )
   {

      // Get the area of the triangle
      float a = C3d::AreaOfTriangle( &x1, &x2, &x3 );
      // Compute the barycentrics.
      // X = U, Y = V, W = Z
      bcoords.x = C3d::AreaOfTriangle( &x2, &x3, &point ) / a;
      bcoords.y = C3d::AreaOfTriangle( &x3, &x1, &point ) / a;
      bcoords.z = C3d::AreaOfTriangle( &x1, &x2, &point ) / a;

      if( 1.0f - ( bcoords.x + bcoords.y + bcoords.z ) - 0.000001f < 0.0f )
      {
         // Calculate the "normalized" barycentrics
         float barytot = bcoords.x + bcoords.y + bcoords.z;
         if( baryover != NULL )
         {
            *baryover = barytot - 1.0f;
         }
         bcoords.x /= barytot;
         bcoords.y /= barytot;
         bcoords.z /= barytot;
      }
      else
      {
         if( baryover != NULL )
         {
            *baryover = 0.0f;
         }
      }

   }

   // Interpolate normal from a point within a triangle
   static void InterpolateNormal( Vertex &point, SVertex *x1, SVertex *x2, SVertex *x3, Vertex &pointNormal )
   {
      Vertex bary;
      float baryover;
      C3d::BarycentricCoords( point, *( x1->v ), *( x2->v ), *( x3->v ), bary, &baryover );
      float baryl = bary.x + bary.y + bary.z;
      pointNormal.x = ( bary.x * x1->normal.x ) + ( bary.y * x2->normal.x ) + ( bary.z * x3->normal.x );
      pointNormal.y = ( bary.x * x1->normal.y ) + ( bary.y * x2->normal.y ) + ( bary.z * x3->normal.y );
      pointNormal.z = ( bary.x * x1->normal.z ) + ( bary.y * x2->normal.z ) + ( bary.z * x3->normal.z );
   }

   // Normalize
   static void Normalize( Vertex &v )
   {
      float veclen = sqrtf( v.x * v.x + v.y * v.y + v.z * v.z );
      if( veclen != 0 )
      {
         v.x /= veclen;
         v.y /= veclen;
         v.z /= veclen;
      }
   }

   // Rotate a point about a center
   static void RotatePoint( Vertex *point, Point3 *center, Vertex *angle )
   {
      float rx = angle->x;
      float ry = angle->y;
      float rz = angle->z;
      float vx = point->x - center->x;
      float vy = point->y - center->y;
      float vz = point->z - center->z;
      CMatrix rmat;
      // Rotation
      //            X                 Y               Z
      /*
      rmat[0][0] =                    cos( ry )  *    cos( rz );
      rmat[0][1] =                                 ( -sin( rz ));
      rmat[0][2] =                    sin( ry );
      rmat[1][0] =                                    sin( rz );
      rmat[1][1] =    cos( rx )                  *    cos( rz );
      rmat[1][2] = ( -sin( rx ));
      rmat[2][0] =                 ( -sin( ry ));
      rmat[2][1] =    sin( rx );
      rmat[2][2] =    cos( rx )  *    cos( ry );
      */
      rmat.BuildRotationXYZ( angle, center );
      point->x = rmat.m[0][0] * vx + rmat.m[1][0] * vy + rmat.m[2][0] * vz + rmat.m[3][0];
      point->y = rmat.m[0][1] * vx + rmat.m[1][1] * vy + rmat.m[2][1] * vz + rmat.m[3][1];
      point->z = rmat.m[0][2] * vx + rmat.m[1][2] * vy + rmat.m[2][2] * vz + rmat.m[3][2];
   }

   static void RotateVertex( Vertex *v, Vertex *center, Vertex *axis, float C )
   {
      float S;
      float a = axis->x, b = axis->y, c = axis->z;
      float x0 = center->x, y0 = center->y, z0 = center->z;
      if( C < 0.0f )
      {
         C = 0.0;
      }
      if( C > 1.0f )
      {
         C = 1.0f;
      }
      S = sqrtf( 1.0f - C * C );
      float x= v->x - x0, y = v->y - y0, z = v->z - z0;
      v->x = (( 1 - C ) * a * a + C     ) * x + (( 1 -C ) * b * a - c * S ) * y + (( 1 - C ) * c * a + b * S ) * z + x0;
      v->y = (( 1 - C ) * a * b + c * S ) * x + (( 1 -C ) * b * b + C     ) * y + (( 1 - C ) * c * b - a * S ) * z + y0;
      v->z = (( 1 - C ) * a * c - b * S ) * x + (( 1 -C ) * b * c + a * S ) * y + (( 1 - C ) * c * c + C     ) * z + z0;
   }


};
//---------------------------------------------------------------------------------------

#endif
