#include "cduvmap.h"
#include <stdio.h>
#include <string.h>

//---------------------------------------------------------------------------------------
// Apply texture coordinates
static void ddhuvmap_go( int argc, const char* argv[] )
{

   if( argc < 10 )
   {
      message_dialog( "Insufficient parameters passed to uvmap plugin" );
      return;
   }

   // UVMap object
   CDUVMap uv;

   // Get params
   int type = 0;
   float stretchx, stretchy, offsetx, offsety, rotate;
   int scaxis = 1;
   if( sscanf( argv[1], "%d", &type ) != 1 )
   {
      message_dialog( "Error parsing UV map type (%s)", argv[1] );
      return;
   }
   if( sscanf( argv[8], "%d", &scaxis ) != 1 )
   {
      message_dialog( "Error parsing Sphere/Cylinder alignment: (%s)", argv[8] );
      return;
   }
   // Only bother here if we're doing global coordinates
   if( type == 11 )
   {
      if( sscanf( argv[2], "%g", &stretchx ) != 1 )
      {
         message_dialog( "Error parsing UV map global coordinates: stretch (x) (%s)", argv[2] );
         return;
      }
      if( sscanf( argv[3], "%g", &stretchy ) != 1 )
      {
         message_dialog( "Error parsing UV map global coordinates: stretch (y) (%s)", argv[3] );
         return;
      }
      if( sscanf( argv[4], "%g", &offsetx ) != 1 )
      {
         message_dialog( "Error parsing UV map global coordinates: shift (x) (%s)", argv[4] );
         return;
      }
      if( sscanf( argv[5], "%g", &offsety ) != 1 )
      {
         message_dialog( "Error parsing UV map global coordinates: shift (y) (%s)", argv[5] );
         return;
      }
      if( sscanf( argv[6], "%g", &rotate ) != 1 )
      {
         message_dialog( "Error parsing UV map global coordinates: rotate (%s)", argv[6] );
         return;
      }
   }

   uv.mUseSS = ( atoi( argv[7] ) != 0 );
   uv.mSpacePoles = ( atoi( argv[9] ) != 0 );

   // Grab selected surfaces
   List *surfs = ac_selection_get_whole_surfaces_all();

   // Undo
   add_undo_texture_coords( "uv map", surfs, NULL );

   // Apply the UV map
   switch( type )
   {
   case 1:
   case 2:
   case 3:
   case 4:
   case 5:
   case 6:
      // Planar
      uv.ApplyPlanar( surfs, type );
      break;
   case 7:
      // Spherical
      uv.ApplySpherical( surfs, scaxis );
      break;
   case 8:
      // Cylindrical
      uv.ApplyCylindrical( surfs, scaxis );
      break;
   case 10:
      // Cylindrical cap
      uv.ApplyCylindricalCap( surfs, scaxis );
      break;
   case 11:
      // Global coordinates
      uv.ApplyGlobal( surfs, stretchx, stretchy, offsetx, offsety, rotate );
      break;
   case 0:
   default:
      // Box
      uv.ApplyBox( surfs );
      break;
   }

   // Remap subdivisions
   if( uv.mSSMap.ContainsSubdivisions() )
   {
      tcl_command( "ac3d remake_subdivisions" );
   }

   // Free the selection list
   list_free( &surfs );

   // Trigger a TCE update
   selection_changed( "uvmap" );

   redraw_all();
   display_status();

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/uvmap.tcl\"", path );
      ac_add_tools_menu_item( "UV map...", "ddhuvmap_dialog",
                              "Activate the UV map plugin" );
      ac_add_command_full( "ddhuvmap_go", ( void * )ddhuvmap_go,
                            0, "argv", "<arg1> <arg2>",
                           "Apply texture coordinates" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "UVMap - version 1.0.1 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "UVMap - Dennis Hawthorne - Applies one of several texture coordinate styles to the selected faces." );
}
//---------------------------------------------------------------------------------------
